# -*- coding: utf-8 -*-
from __future__ import (absolute_import, division, print_function)

from ..einstein_smoluchowski import electrical_mobility_from_D
from ..units import default_units, allclose, default_constants
from ..util.testing import requires
from ..units import units_library


def test_electrical_mobility_from_D():
    D = 3
    z = -2
    T = 100
    mu = electrical_mobility_from_D(D, z, T)
    e = 1.60217657e-19
    kB = 1.3806488e-23
    ref = z*e*D/(kB*T)
    assert allclose(mu, ref, rtol=1e-5)

    mu2 = electrical_mobility_from_D(3, -2, 100)
    assert allclose(mu2, -2*1.60217657e-19*3/1.3806488e-23/100, rtol=1e-5)


@requires(units_library)
def test_electrical_mobility_from_D__with_units():
    metre = default_units.metre
    second = default_units.second
    kelvin = default_units.kelvin
    coulomb = default_units.coulomb
    kilogram = default_units.kilogram

    D = 3*metre/second**2
    z = -2
    T = 100*kelvin
    mu = electrical_mobility_from_D(D, z, T, default_constants, default_units)
    e = 1.60217657e-19 * coulomb
    kB = 1.3806488e-23 * metre**2 * kilogram * second**-2 / kelvin
    ref = z*e*D/(kB*T)
    assert allclose(mu, ref, rtol=1e-5)

    mu2 = electrical_mobility_from_D(3, -2, 100)
    assert allclose(mu2, -2*1.60217657e-19*3/1.3806488e-23/100, rtol=1e-5)



# -*- coding: utf-8 -*-
from __future__ import (absolute_import, division, print_function)

from collections import OrderedDict
from ..chemistry import Substance
from .numbers import number_to_scientific_html


class Table(object):
    def __init__(self, rows, headers=None):
        self.rows, self.headers = rows, headers

    def _html(self, printer, **kwargs):
        def map_fmt(cont, fmt, joiner='\n'):
            return joiner.join(map(lambda x: fmt % printer._print(x, **kwargs), cont))
        rows = (
            [map_fmt(self.headers, '<th>%s</th>')] +
            [map_fmt(row, '<td>%s</td>') for row in self.rows]
        )
        return '<table>%s</table>' % map_fmt(rows, '<tr>%s</tr>')

    def _repr_html_(self):
        from .web import html
        return html(self)


def as_per_substance_html_table(cont, substances=None, header=None,
                                substance_factory=Substance.from_formula):
    """ """
    if substances is None:
        substances = OrderedDict([(k, substance_factory(k)) for k in cont])

    def _elem(k):
        try:
            return cont[k]
        except (IndexError, TypeError):
            return cont[list(substances.keys()).index(k)]
    rows = [(v.html_name, number_to_scientific_html(_elem(k))) for k, v in substances.items()]
    return Table(rows, ['Substance', header or ''])



# -*- coding: utf-8 -*-
from __future__ import (absolute_import, division, print_function)

from .util.pyutil import ChemPyDeprecationWarning

import warnings

from .kinetics.eyring import (
    eyring_equation, EyringParam, EyringParamWithUnits
)


warnings.warn("use .kinetics.eyring instead of .eyring", ChemPyDeprecationWarning)



# -*- coding: utf-8 -*-
from __future__ import (absolute_import, division, print_function)

import math
from operator import add
from functools import reduce

import pytest

from chempy import Substance
from chempy.units import (
    allclose, units_library, default_constants, Backend, to_unitless,
    SI_base_registry, default_units as u
)
from ..testing import requires
from ..pyutil import defaultkeydict
from .._expr import Expr, mk_Poly, mk_PiecewisePoly, create_Piecewise, create_Poly, Log10
from ..parsing import parsing_library


class HeatCapacity(Expr):
    parameter_keys = ('temperature',)


class EinsteinSolid(HeatCapacity):
    parameter_keys = HeatCapacity.parameter_keys + ('molar_gas_constant',)
    argument_names = ('einstein_temperature', 'molar_mass')

    def __call__(self, variables, backend=math):
        TE, molar_mass = self.all_args(variables, backend=backend)  # einstein_temperature
        T, R = self.all_params(variables, backend=backend)
        # Canonical ensemble:
        molar_c_v = 3*R*(TE/(2*T))**2 * backend.sinh(TE/(2*T))**-2
        return molar_c_v/molar_mass


def _get_cv(kelvin=1, gram=1, mol=1):

    Al = Substance.from_formula('Al', data={'DebyeT': 428*kelvin})
    Be = Substance.from_formula('Be', data={'DebyeT': 1440*kelvin})

    def einT(s):
        return 0.806*s.data['DebyeT']
    return {s.name: EinsteinSolid([einT(s), s.mass * gram/mol]) for s in (Al, Be)}


@requires(parsing_library)
def test_Expr():
    cv = _get_cv()
    _ref = 0.8108020083055849
    assert abs(cv['Al']({'temperature': 273.15, 'molar_gas_constant': 8.3145}) - _ref) < 1e-14


def _poly(args, x, backend=None):
    x0, coeffs = args[0], args[1:]
    return reduce(add, [c*(x-x0)**i for i, c in enumerate(coeffs)])


@requires(parsing_library)
def test_Expr__nested_Expr():
    Poly = Expr.from_callback(_poly, parameter_keys=('x',), argument_names=('x0', Ellipsis))
    T = Poly([3, 7, 5])

    cv = _get_cv()
    _ref = 0.8108020083055849
    args = {'temperature': T, 'x': (273.15-7)/5 + 3, 'molar_gas_constant': 8.3145}
    assert abs(cv['Al'](args) - _ref) < 1e-14
    Al2 = cv['Al']/2
    assert abs(Al2(args) - _ref/2) < 1e-14


def test_nargs():
    class A(Expr):
        nargs = 1

    with pytest.raises(ValueError):
        A([1, 2])


@requires('sympy')
def test_Expr_symbolic():
    import sympy
    cv = _get_cv()
    R, T = sympy.symbols('R T')
    sexpr = cv['Be']({'temperature': T, 'molar_gas_constant': R}, backend=sympy)
    assert sexpr.free_symbols == set([T, R])


@requires(units_library)
def test_Expr_units():
    cv = _get_cv(u.kelvin, u.gram, u.mol)
    R = default_constants.molar_gas_constant.rescale(u.joule/u.mol/u.kelvin)

    def _check(T=273.15*u.kelvin):
        result = cv['Be']({'temperature': T, 'molar_gas_constant': R}, backend=Backend())
        ref = 0.7342617587256584*u.joule/u.gram/u.kelvin
        assert abs(to_unitless((result - ref)/ref)) < 1e-10
    _check()
    _check(491.67*u.rankine)


@requires(units_library)
def test_Expr_dedimensionalisation__1():
    cv = _get_cv(u.kelvin, u.gram, u.mol)
    units, expr = cv['Be'].dedimensionalisation(SI_base_registry)
    assert units == [u.kelvin, u.kg/u.mol]
    assert abs(expr.args[0] - 0.806*1440) < 1e-14
    assert abs(expr.args[1] - 9.01218e-3) < 1e-7


@requires(units_library)
def test_Expr_dedimensionalisation__2():
    Poly = Expr.from_callback(_poly, parameter_keys=('E',), argument_names=('x0', Ellipsis))
    T = Poly([3*u.J, 7*u.K, 5*u.K/u.J])
    T = Poly([0.7170172084130019*u.cal, 12.6*u.Rankine, 5*u.K/u.J])

    _ref = 0.8108020083055849  # Al at 273.15 K with R=8.3145
    cv_Al = _get_cv(u.kelvin, u.gram, u.mol)['Al']

    assert isinstance(cv_Al, EinsteinSolid)
    assert cv_Al.args[0] == 0.806*428*u.kelvin
    assert abs(cv_Al({'temperature': 273.15*u.K, 'molar_gas_constant': 8.3145*u.J/u.K/u.mol}) -
               _ref*u.J/u.gram/u.kelvin) < 1e-14

    cv_Al_units, Al_dedim = cv_Al.dedimensionalisation(SI_base_registry)
    assert allclose(cv_Al_units, [u.K, u.kg/u.mol])
    assert isinstance(Al_dedim, EinsteinSolid)
    T_units, dT = T.dedimensionalisation(SI_base_registry)
    assert allclose(T_units, [u.J, u.K, u.K/u.J])
    assert allclose(dT.args, [3, 7, 5])
    assert abs(Al_dedim({'temperature': 273.15, 'molar_gas_constant': 8.3145}) - _ref*1000) < 1e-14
    assert abs(Al_dedim({'temperature': dT, 'E': (273.15-7)/5 + 3, 'molar_gas_constant': 8.3145}) - _ref*1000) < 1e-14


@requires(units_library)
def test_Expr_dedimensionalisation__nested():
    Poly = Expr.from_callback(_poly, parameter_keys=('E',), argument_names=('x0', Ellipsis))
    TE = Poly([3*u.J, 7*u.K, 5*u.K/u.J])
    TE = Poly([0.7170172084130019*u.cal, 12.6*u.Rankine, 5*u.K/u.J]) * 0.806*428/273.15

    _ref = 0.8108020083055849  # Al at 273.15 K with R=8.3145
    cv_Al = _get_cv(u.kelvin, u.gram, u.mol)['Al']
    cv_Al_units, Al_dedim = cv_Al.dedimensionalisation(SI_base_registry, {'einstein_temperature': TE})
    assert abs(Al_dedim({'temperature': 273.15, 'E': (273.15-7)/5 + 3, 'molar_gas_constant': 8.3145}) -
               _ref*1000) < 1e-14


def test_Expr__from_callback():
    def two_dim_gauss(args, x, y, backend=None):
        A, x0, y0, sx, sy = args
        xp, yp = x-x0, y-y0
        vx, vy = 2*sx**2, 2*sy**2
        return A*backend.exp(-(xp**2/vx + yp**2/vy))

    TwoDimGauss = Expr.from_callback(two_dim_gauss, parameter_keys=('x', 'y'), nargs=5)
    with pytest.raises(ValueError):
        TwoDimGauss([1, 2])
    args = [3, 2, 1, 4, 5]
    g1 = TwoDimGauss(args)
    ref = two_dim_gauss(args, 6, 7, math)
    assert abs(g1({'x': 6, 'y': 7}) - ref) < 1e-14


def test_mk_Poly():
    Poly = mk_Poly('T', reciprocal=True)
    p = Poly([3, 2, 5, 7, 8, 2, 9])
    assert p.eval_poly({'T': 13}) == 2.57829
    assert p.parameter_keys == ('T',)


def test_Expr__nargs():

    class Linear(Expr):
        """ Arguments: p0, p1 """
        nargs = 2
        parameter_keys = ('x',)

        def __call__(self, variables, backend=None):
            p0, p1 = self.all_args(variables)
            return p0 + p1*variables['x']

    l1 = Linear([3, 2])
    assert l1(dict(x=5)) == 13
    with pytest.raises(ValueError):
        Linear([3])
    with pytest.raises(ValueError):
        Linear([3, 2, 1])

    l2 = Linear([3, 2], ['a', 'b'])
    # with pytest.raises(ValueError):
    #     Linear([3, 2], ['a'])
    with pytest.raises(ValueError):
        Linear([3, 2], ['a', 'b', 'c'])

    assert l2(dict(x=5)) == 13
    assert l2(dict(x=5, a=11, b=13)) == 11 + 13*5


def test_PiecewisePoly():
    Poly = mk_Poly('temperature')

    p1 = Poly([0, 1, 0.1])
    assert p1.eval_poly({'temperature': 10}) == 2

    p2 = Poly([0, 3, -.1])
    assert p2.eval_poly({'temperature': 10}) == 2

    TPiecewisePoly = mk_PiecewisePoly('temperature')
    tpwp = TPiecewisePoly.from_polynomials([(0, 10), (10, 20)], [p1, p2])
    assert tpwp.eval_poly({'temperature': 5}) == 1.5
    assert tpwp.eval_poly({'temperature': 15}) == 1.5
    assert tpwp.parameter_keys == ('temperature',)

    with pytest.raises(ValueError):
        tpwp.eval_poly({'temperature': 21})


def test_create_Piecewise_Poly():
    PolyT = create_Poly('Tmpr')

    p1 = PolyT([1, 0.1])
    assert p1({'Tmpr': 10}) == 2

    p2 = PolyT([3, -.1])
    assert p2({'Tmpr': 10}) == 2

    PiecewiseT = create_Piecewise('Tmpr')
    pw = PiecewiseT([0, p1, 10, p2, 20])
    assert pw({'Tmpr': 5}) == 1.5
    assert pw({'Tmpr': 15}) == 1.5
    assert pw.parameter_keys == ('Tmpr',)

    with pytest.raises(ValueError):
        pw({'Tmpr': 21})


@requires('sympy')
def test_PiecewisePoly__sympy():
    import sympy as sp
    Poly = mk_Poly('T')
    p1 = Poly([0, 1, 0.1])
    p2 = Poly([0, 3, -.1])

    TPiecewisePoly = mk_PiecewisePoly('temperature')
    tpwp = TPiecewisePoly([2, 2, 0, 10, 2, 10, 20, 0, 1, 0.1, 0, 3, -.1])
    x = sp.Symbol('x')
    res = tpwp.eval_poly({'temperature': x}, backend=sp)
    assert isinstance(res, sp.Piecewise)
    assert res.args[0][0] == 1 + 0.1*x
    assert res.args[0][1] == sp.And(0 <= x, x <= 10)
    assert res.args[1][0] == 3 - 0.1*x
    assert res.args[1][1] == sp.And(10 <= x, x <= 20)

    with pytest.raises(ValueError):
        tpwp.from_polynomials([(0, 10), (10, 20)], [p1, p2])


@requires('sympy')
def test_create_Piecewise_Poly__sympy():
    import sympy as sp
    Poly = create_Poly('Tmpr')
    p1 = Poly([1, 0.1])
    p2 = Poly([3, -.1])

    TPw = create_Piecewise('Tmpr')
    pw = TPw([0, p1, 10, p2, 20])
    x = sp.Symbol('x')
    res = pw({'Tmpr': x}, backend=sp)
    assert isinstance(res, sp.Piecewise)
    assert res.args[0][0] == 1 + 0.1*x
    assert res.args[0][1] == sp.And(0 <= x, x <= 10)
    assert res.args[1][0] == 3 - 0.1*x
    assert res.args[1][1] == sp.And(10 <= x, x <= 20)


@requires('sympy')
def test_create_Piecewise__nan_fallback__sympy():
    import sympy as sp

    TPw = create_Piecewise('Tmpr', nan_fallback=True)
    pw = TPw([0, 42, 10, 43, 20])
    x = sp.Symbol('x')
    res = pw({'Tmpr': x}, backend=sp)
    assert isinstance(res, sp.Piecewise)
    assert res.args[0][0] == 42
    assert res.args[0][1] == sp.And(0 <= x, x <= 10)
    assert res.args[1][0] == 43
    assert res.args[1][1] == sp.And(10 <= x, x <= 20)
    assert res.args[2][0].name.lower() == 'nan'
    assert res.args[2][1] == True  # noqa


def test_BinaryExpr():
    Poly = Expr.from_callback(_poly, parameter_keys=('x',), argument_names=('x0', Ellipsis))
    p1 = Poly([1, 2, 3])
    p2 = Poly([2, 3, 4])
    assert p1({'x': 5}) == 14
    assert p2({'x': 5}) == 15
    assert (p1+p2)({'x': 5}) == 14+15
    assert (p1-p2)({'x': 5}) == 14-15
    assert (p1*p2)({'x': 5}) == 14*15
    assert (p1/p2)({'x': 5}) == 14/15

    assert (p1+2)({'x': 5}) == 14+2
    assert (p1-2)({'x': 5}) == 14-2
    assert (p1*2)({'x': 5}) == 14*2
    assert (p1/2)({'x': 5}) == 14/2

    assert (2+p1)({'x': 5}) == 2+14
    assert (2-p1)({'x': 5}) == 2-14
    assert (2*p1)({'x': 5}) == 2*14
    assert (2/p1)({'x': 5}) == 2/14

    assert p1 + 0 == p1
    assert p1 * 1 == p1
    assert p1 + p2 == p1 + p2
    assert p1 + p2*1 == p1 + p2 + 0

    assert -(-p1) == p1


@pytest.mark.slow
@requires('sympy')
def test_Expr__latex():
    Poly = Expr.from_callback(_poly, parameter_keys=('x',), argument_names=('x0', Ellipsis))
    p = Poly([1, 2, 3, 4])
    import sympy
    t = sympy.Symbol('t')
    ref = sympy.latex((2 + 3*(t-1) + 4*(t-1)**2).simplify())
    assert p.latex({'x': 't'}) == ref

    TE = Poly([3, 7, 5])
    cv_Al = _get_cv()['Al']
    T, E, R, m = sympy.symbols('T E R m')
    _TE = 7 + 5*(E-3)
    ref = sympy.latex(((3*R*(_TE/(2*T))**2 * sympy.sinh(_TE/(2*T))**-2)/m).simplify())
    cv_Al.unique_keys = ('TE_Al', 'm_Al')
    res = cv_Al.latex({'TE_Al': TE, 'temperature': 'T', 'x': 'E', 'molar_gas_constant': 'R', 'm_Al': 'm'})
    assert res == ref

    X = sympy.symbols('X')
    _TE2 = 7 + 5*(X-3)
    ref2 = sympy.latex(((3*R*(_TE2/(2*T))**2 * sympy.sinh(_TE2/(2*T))**-2)/m).simplify())
    res2 = cv_Al.latex({'TE_Al': TE, 'temperature': 'T', 'molar_gas_constant': 'R', 'm_Al': 'm'},
                       default=lambda k: k.upper())
    assert res2 == ref2


class Pressure1(Expr):
    argument_names = ('n',)
    parameter_keys = ('temperature', 'volume', 'R')

    def __call__(self, variables, backend=None):
        n, = self.all_args(variables, backend=backend)
        T, V, R = self.all_params(variables, backend=backend)
        return n*R*T/V


def test_Expr__single_arg():
    p = Pressure1(3)
    assert abs(p({'temperature': 273.15, 'volume': 0.17, 'R': 8.314}) - 3*8.314*273.15/0.17) < 1e-15


@requires(units_library)
def test_Expr__single_arg__units():
    p = Pressure1(3*u.mol)
    variables = {'temperature': 273.15*u.kelvin, 'volume': 170*u.dm3, 'R': 8.314*u.J/u.K/u.mol}
    assert allclose(p(variables), 3*8.314*273.15/0.17*u.Pa)


class Pressure2(Pressure1):
    def args_dimensionality(self):
        return ({'amount': 1},)


@requires(units_library)
def test_Expr__single_arg__units__dimensionality():
    p = Pressure2(unique_keys=('n1',))
    variables = {'temperature': 273.15*u.kelvin, 'volume': 170*u.dm3, 'R': 8.314*u.J/u.K/u.mol}
    assert allclose(p(dict(n1=3*u.mol, **variables)), 3*8.314*273.15/0.17*u.Pa)


def test_Expr__argument_defaults():

    class MyExpr(Expr):
        argument_names = ('a', 'b', 'c')
        argument_defaults = (17, 23)

        def __call__(self, variables={}, backend=math):
            a, b, c = self.all_args(variables, backend=backend)
            return a*b*c

    assert MyExpr([15])() == 15*17*23
    assert MyExpr([15, 17])() == 15*17*23
    assert MyExpr([15, 19])() == 15*19*23
    assert MyExpr([15, 19, 29])() == 15*19*29
    assert MyExpr(dict(zip('abc', [15, 19, 29])))() == 15*19*29


class MyK(Expr):
    argument_names = ('H', 'S')
    parameter_keys = ('T',)
    R = 8.3145

    def __call__(self, variables, backend=math):
        H, S = self.all_args(variables, backend=backend)
        T, = self.all_params(variables, backend=backend)
        return backend.exp(-(H - T*S)/(self.R*T))


def test_Expr__no_args():
    K1 = MyK(unique_keys=('H1', 'S1'))
    K2 = MyK(unique_keys=('H2', 'S2'))
    add = K1 + K2
    T = 298.15
    res = add({'H1': 2, 'H2': 3, 'S1': 5, 'S2': 7, 'T': T})
    RT = 8.3145 * 298.15
    ref = math.exp(-(2 - T*5)/RT) + math.exp(-(3 - T*7)/RT)
    assert abs(res - ref) < 1e-14


@requires(units_library)
def test_Expr__equality():
    K1 = MyK(unique_keys=('H1', 'S1'))
    K2 = MyK(unique_keys=('H2', 'S2'))
    assert K1 != K2
    assert K1 == K1
    K3 = MyK([23e3*u.J/u.mol, 42*u.J/u.mol/u.K])
    K4 = MyK([23e3, 42])
    K5 = MyK([24e3*u.J/u.mol, 42*u.J/u.mol/u.K])
    K6 = MyK([23e3, 43])
    assert K3 == K3
    assert K4 == K4
    assert K5 != K3
    assert K4 != K6
    assert K3 != K4


@requires('sympy')
def test_Expr__no_args__symbolic():
    K1 = MyK(unique_keys=('H1', 'S1'))
    K2 = MyK(unique_keys=('H2', 'S2'))
    add = K1 + K2
    import sympy
    v = defaultkeydict(sympy.Symbol)
    res = add(v, backend=sympy)
    R = 8.3145
    expr1 = sympy.exp(-(v['H1'] - v['T']*v['S1'])/R/v['T'])
    expr2 = sympy.exp(-(v['H2'] - v['T']*v['S2'])/R/v['T'])
    ref = expr1 + expr2
    assert (res - ref).simplify() == 0


class MyK2(Expr):
    argument_names = ('H', 'S', 'Cp', 'Tref')
    argument_defaults = (0, 298.15)
    parameter_keys = ('T')
    R = 8.3145

    def __call__(self, variables, backend=math):
        H, S, Cp, Tref = self.all_args(variables, backend=backend)
        T, = self.all_params(variables, backend=backend)
        _H = H + Cp*(T-Tref)
        _S = S + Cp*backend.log(T/Tref)
        return backend.exp(-(_H - T*_S)/(self.R*T))


def test_Expr__no_args__arg_defaults():
    K1 = MyK2(unique_keys=('H1', 'S1', 'Cp1'))
    K2 = MyK2(unique_keys=('H2', 'S2'))
    add = K1 + K2
    assert add.all_unique_keys() == set(['H1', 'H2', 'S1', 'S2', 'Cp1'])

    T = 293.15
    res = add({'H1': 2, 'H2': 3, 'S1': 5, 'S2': 7, 'T': T, 'Cp1': 13})
    RT = 8.3145 * T
    H1p = 2 + 13*(T - 298.15)
    S1p = 5 + 13*math.log(T/298.15)
    ref = math.exp(-(H1p - T*S1p)/RT) + math.exp(-(3 - T*7)/RT)
    assert abs(res - ref) < 1e-14


def test_create_Piecewise():
    PW = create_Piecewise('T')
    Ha, Sa, Hb, Sb, Ta, Tb = 40e3, -60, 37e3, -42, 293.15, 303.15
    a = MyK([Ha, Sa])
    b = MyK([Hb, Sb])
    pw = PW([273.15, a, 298.15, b, 323.15])  # 0, 25, 50 *C
    res_a = pw({'T': Ta})  # 20 *C
    res_b = pw({'T': Tb})  # 30 *C
    ref_a = math.exp(-(Ha - Ta*Sa)/(MyK.R*Ta))
    ref_b = math.exp(-(Hb - Tb*Sb)/(MyK.R*Tb))
    assert abs(res_a - ref_a) < 1e-14
    assert abs(res_b - ref_b) < 1e-14


def test_create_Poly():
    PolyT = create_Poly('T')
    p = PolyT([1, 2, 3, 4, 5])
    assert p({'T': 11}) == 1 + 2*11 + 3*11**2 + 4*11**3 + 5*11**4


def test_pow():
    PolyT = create_Poly('T')
    p = PolyT([1, 2, 3])
    p_to_two = p**2
    res = p_to_two({'T': 3})
    ref = (1 + 2*3 + 3*9)**2
    assert abs(res - ref) < 1e-12
    two_to_p = 2**p
    res = two_to_p({'T': 3})
    ref = 2**(1 + 2*3 + 3*9)
    assert abs(res - ref) < 1e-12


def test_functions():
    PolyT = create_Poly('T')
    p = PolyT([1, 2, 3])
    lgp = Log10(p)
    res = lgp({'T': 3})
    ref = math.log10(1 + 2*3 + 3*9)
    assert abs(res - ref) < 1e-12


def test_str_arg():
    PolyT = create_Poly('x')
    p = PolyT([1, 2, 3])
    x = Log10('T')
    res = p({'x': x, 'T': 1000})
    ref = 1 + 2*3 + 3*9
    assert abs(res - ref) < 1e-12


def test_implicit_str():
    PolyT = create_Poly('x')
    p = PolyT([1, 2, 3])
    expr1 = p / 'u'
    assert abs(expr1({'x': 3, 'u': 5}) - (1 + 2*3 + 3*9)/5) < 1e-12
    expr2 = 'u' / p
    assert abs(expr2({'x': 3, 'u': 5}) - 5/(1 + 2*3 + 3*9)) < 1e-12
    assert expr1.all_parameter_keys() == set(['x'])
    assert expr2.all_parameter_keys() == set(['x'])



# -*- coding: utf-8 -*-
from __future__ import (absolute_import, division, print_function)

try:
    import numpy as np
except ImportError:
    np = None
else:
    from ..stoich import (
        get_coeff_mtx, decompose_yields,
    )
from chempy.units import units_library, allclose, _sum
from ..testing import requires


@requires('numpy')
def test_get_coeff_mtx():
    r = [
        ({'A': 1},       {'B': 1}),
        ({'A': 1, 'B': 1}, {'C': 2})
    ]
    A = get_coeff_mtx('ABC', r)
    Aref = np.array([
        [-1, -1],
        [1, -1],
        [0,  2]
    ])
    assert np.allclose(A, Aref)


@requires('numpy')
def test_decompose_yields_1():
    from chempy import Reaction

    gamma_yields = {
        'OH-': 0.5,
        'H2O2': 0.7,
        'OH': 2.7,
        'H2': 0.45,
        'H': 0.66,
        'H+': 3.1,
        'HO2': 0.02,
        'e-(aq)': 2.6,
    }

    rxns = [
        Reaction({'H2O': 1}, {'H+': 1, 'OH-': 1}),
        Reaction({'H2O': 1}, {'H+': 1, 'e-(aq)': 1, 'OH': 1}),
        Reaction({'H2O': 1}, {'H': 2, 'H2O2': 1}, inact_reac={'H2O': 1}),
        Reaction({'H2O': 1}, {'H2': 1, 'H2O2': 1}, inact_reac={'H2O': 1}),
        Reaction({'H2O': 1}, {'H2': 1, 'OH': 2}, inact_reac={'H2O': 1}),
        Reaction({'H2O': 1}, {'H2': 3, 'HO2': 2}, inact_reac={'H2O': 3}),
    ]

    k = decompose_yields(gamma_yields, rxns)
    k_ref = [0.5, 2.6, 0.33, 0.37, 0.05, 0.01]

    assert np.allclose(k, k_ref)

    G_H2O = sum(rxn.net_stoich(['H2O'])[0]*k[i] for
                i, rxn in enumerate(rxns))

    assert abs(G_H2O+4.64) < 1e-3


@requires(units_library)
def test_decompose_yields__units_1():
    from chempy import Reaction
    from chempy.units import default_units as u
    gamma_yields = {
        'OH-': 0.5*u.per100eV,
        'H2O2': 0.7*u.per100eV,
        'OH': 2.7*u.per100eV,
        'H2': 0.45*u.per100eV,
        'H': 0.66*u.per100eV,
        'H+': 3.1*u.per100eV,
        'HO2': 0.02*u.per100eV,
        'e-(aq)': 2.6*u.per100eV,
    }

    rxns = [
        Reaction({'H2O': 1}, {'H+': 1, 'OH-': 1}),
        Reaction({'H2O': 1}, {'H+': 1, 'e-(aq)': 1, 'OH': 1}),
        Reaction({'H2O': 1}, {'H': 2, 'H2O2': 1}, inact_reac={'H2O': 1}),
        Reaction({'H2O': 1}, {'H2': 1, 'H2O2': 1}, inact_reac={'H2O': 1}),
        Reaction({'H2O': 1}, {'H2': 1, 'OH': 2}, inact_reac={'H2O': 1}),
        Reaction({'H2O': 1}, {'H2': 3, 'HO2': 2}, inact_reac={'H2O': 3}),
    ]

    k = decompose_yields(gamma_yields, rxns)
    k_ref = [0.5, 2.6, 0.33, 0.37, 0.05, 0.01]*u.per100eV

    assert allclose(k, k_ref)

    G_H2O = [rxn.net_stoich(['H2O'])[0]*k[i] for i, rxn in enumerate(rxns)]
    ref = 4.64*u.per100eV
    assert abs((_sum(G_H2O)+ref)/ref) < 1e-3


@requires('numpy')
def test_decompose_yields_2():
    from chempy import Reaction
    yields = {'B': 3.0, 'C': 24.0}
    rxns = [
        Reaction({'A': 1}, {'B': 1, 'C': 1}, inact_reac={'A': 1}),
        Reaction({'A': 1}, {'C': 3})
    ]
    k = decompose_yields(yields, rxns)
    k_ref = [3, 7]

    rtol = 1e-12
    for a, b in zip(k, k_ref):
        assert abs(a-b) < abs(a*rtol)



# -*- coding: utf-8 -*-
"""
General utilities and exceptions.
"""
from __future__ import (absolute_import, division, print_function)

from collections import defaultdict, namedtuple, Mapping, OrderedDict
from functools import wraps
from itertools import product
import os
import warnings

from .. import __url__
from .deprecation import Deprecation


class NoConvergence(Exception):
    pass


class ChemPyDeprecationWarning(DeprecationWarning):
    pass


def deprecated(*args, **kwargs):
    """ Helper to :class:`Deprecation` for using ChemPyDeprecationWarning. """
    return Deprecation(
        *args, issues_url=lambda s: __url__ + '/issues/' +
        s.lstrip('gh-'), warning=ChemPyDeprecationWarning, **kwargs)

warnings.simplefilter(os.environ.get('CHEMPY_DEPRECATION_FILTER', 'once'),
                      ChemPyDeprecationWarning)


class DeferredImport(object):
    def __init__(self, modname, arg=None, decorators=None):
        self._modname = modname
        self._arg = arg
        self._decorators = decorators
        self._cache = None

    @property
    def cache(self):
        if self._cache is None:
            if self._arg is None:
                obj = __import__(self._modname)
            else:
                obj = getattr(__import__(self._modname, globals(), locals(), [self._arg]), self._arg)
            if self._decorators is not None:
                for deco in self._decorators:
                    obj = deco(obj)
            self._cache = obj
        return self._cache

    def __getattribute__(self, attr):
        if attr in ('_modname', '_arg', '_cache', 'cache', '_decorators'):
            return object.__getattribute__(self, attr)
        else:
            return getattr(self.cache, attr)

    def __call__(self, *args, **kwargs):
        return self.cache(*args, **kwargs)


class NameSpace:
    """ Used to wrap, e.g. modules.

    Parameters
    ----------
    default : module
        The underlying module. Acts as a fallback for attribute access.

    Examples
    --------
    >>> import numpy
    >>> my_numpy = NameSpace(numpy)
    >>> my_numpy.array = lambda *args, **kwargs: list(numpy.array(*args, **kwargs))
    >>> isinstance(my_numpy.array([2, 3]), list)
    True
    >>> isinstance(numpy.array([2, 3]), list)
    False

    """

    def __init__(self, default):
        self._NameSpace_default = default
        self._NameSpace_attr_store = {}

    def __getattr__(self, attr):
        if attr.startswith('_NameSpace_'):
            return self.__dict__[attr]
        else:
            try:
                return self._NameSpace_attr_store[attr]
            except KeyError:
                return getattr(self._NameSpace_default, attr)

    def __setattr__(self, attr, val):
        if attr.startswith('_NameSpace_'):
            self.__dict__[attr] = val
        else:
            self._NameSpace_attr_store[attr] = val

    def as_dict(self):
        items = self._NameSpace_default.__dict__.items()
        result = {k: v for k, v in items if not k.startswith('_')}
        result.update(self._NameSpace_attr_store)
        return result


class AttributeContainer(object):
    """ Used to turn e.g. a dictionary to a module-like object.

    Parameters
    ----------
    d : dict

    Examples
    --------
    >>> def RT(T, const):
    ...     return T*const.molar_gas_constant
    ...
    >>> from quantities import constants
    >>> RT(273.15, constants)
    array(273.15) * R
    >>> my_constants = AttributeContainer(molar_gas_constant=42)
    >>> RT(273.15, my_constants)
    11472.3

    """
    def __init__(self, **kwargs):
        self.__dict__.update(**kwargs)

    def as_dict(self):
        return self.__dict__.copy()


class defaultkeydict(defaultdict):
    """ defaultdict where default_factory should have the signature key -> value

    Examples
    --------
    >>> d = defaultkeydict(lambda k: '[%s]' % k, {'a': '[a]', 'b': '[B]'})
    >>> d['a']
    '[a]'
    >>> d['b']
    '[B]'
    >>> d['c']
    '[c]'

    """

    def __missing__(self, key):
        if self.default_factory is None:
            raise KeyError("Missing key: %s" % key)
        else:
            self[key] = self.default_factory(key)
        return self[key]


def defaultnamedtuple(typename, field_names, defaults=()):
    """ Generates a new subclass of tuple with default values.

    Parameters
    ----------
    typename : string
        The name of the class.
    field_names : str or iterable
        An iterable of splitable string.
    defaults : iterable
        Default values for ``field_names``, counting ``[-len(defaults):]``.

    Examples
    --------
    >>> Body = defaultnamedtuple('Body', 'x y z density', (1.0,))
    >>> Body.__doc__
    'Body(x, y, z, density)'
    >>> b = Body(10, z=3, y=5)
    >>> b._asdict()
    OrderedDict([('x', 10), ('y', 5), ('z', 3), ('density', 1.0)])

    Returns
    -------
    A new tuple subclass named ``typename``

    """
    Tuple = namedtuple(typename, field_names)
    Tuple.__new__.__defaults__ = (None,) * len(Tuple._fields)
    if isinstance(defaults, Mapping):
        Tuple.__new__.__defaults__ = tuple(Tuple(**defaults))
    else:
        nmissing = len(Tuple._fields) - len(defaults)
        defaults = (None,)*nmissing + tuple(defaults)
        Tuple.__new__.__defaults__ = tuple(Tuple(*defaults))
    return Tuple


def multi_indexed_cases(od):
    """ Returns a list of length-2 tuples

    Each tuple consist of a multi-index (tuple of integers) and a dictionary.

    Parameters
    ----------
    od : OrderedDict
        Maps each key to a number of values. ``list`` and ``tuple`` instances
        are converted to ``OrderedDict``.

    Examples
    --------
    >>> from chempy.util.pyutil import multi_indexed_cases
    >>> cases = multi_indexed_cases([('a', [1, 2, 3]), ('b', [False, True])])
    >>> len(cases)
    6
    >>> midxs, case_kws = zip(*cases)
    >>> midxs[0]
    (0, 0)
    >>> case_kws[0] == {'a': 1, 'b': False}
    True

    Returns
    -------
    List of length-2 tuples, each consisting of one tuple of inidices and one dictionary.

    """
    if isinstance(od, (list, tuple)):
        od = OrderedDict(od)
    if not isinstance(od, OrderedDict):
        raise NotImplementedError("Expected an OrderedDcit")
    keys, values = map(list, [od.keys(), od.values()])
    return [(mi, {k: v[i] for k, v, i in zip(keys, values, mi)}) for mi in product(*map(range, map(len, values)))]


def memoize(max_nargs=0):
    def decorator(func):
        @wraps(func)
        def wrapper(*args):
            if max_nargs is not None and len(args) > max_nargs:
                raise ValueError("memoization error")
            if args not in wrapper.results:
                wrapper.results[args] = func(*args)
            return wrapper.results[args]
        wrapper.results = {}
        return wrapper
    return decorator



# -*- coding: utf-8 -*-
"""
Convenience functions for presenting reaction systems in tables.
"""
from __future__ import (absolute_import, division, print_function)


import os
import shutil
import subprocess
import tempfile

from ..printing import latex
from ..kinetics.rates import RadiolyticBase
from ..units import to_unitless, get_derived_unit

tex_templates = {
    'document': {
        'default': r"""
\documentclass[a4paper,9pt]{article}
\pagestyle{empty}
\usepackage[paper=a4paper,margin=1cm]{geometry}
%(usepkg)s
\hypersetup{
  bookmarksnumbered=true,
  breaklinks=false,
  raiselinks=true,
  pdfborder={0 0 0},
  colorlinks=true,
  plainpages=false,
  pdfstartview={FitH},
  pdfcreator={LaTeX with hyperref package},
  citecolor=teal,
  linkcolor=red,
  urlcolor=blue,
}
\begin{document}
%(begins)s
%(table)s
%(ends)s
\end{document}
"""
    },
    'table': {
        'default': r"""
\begin{%(table_env)s}
\centering
\label{tab:%(label)s}
\caption[%(short_cap)s]{%(long_cap)s}
\begin{tabular}{%(alignment)s}
\toprule
%(header)s
\midrule
%(body)s
\bottomrule
\end{tabular}
\end{%(table_env)s}""",
        'longtable': r"""
\begin{%(table_env)s}{%(alignment)s}
\caption[%(short_cap)s]{%(long_cap)s
\label{tab:%(label)s}}\\
\toprule
%(header)s
\midrule
%(body)s
\bottomrule
\end{%(table_env)s}"""
    }
}


def render_tex_to_pdf(contents, texfname, pdffname, output_dir, save):
    """ Generates a pdf from a tex file by calling pdflatex

    Parameters
    ----------
    contents : str
    texfname : path
    pdffname : path
    output_dir : path
    save : path or bool or str(bool)

    """
    created_tempdir = False
    try:
        if output_dir is None:
            output_dir = tempfile.mkdtemp()
            created_tempdir = True
        texpath = os.path.join(output_dir, texfname)
        pdfpath = os.path.join(output_dir, pdffname)
        cmds = ['pdflatex', '-halt-on-error', '-interaction',
                'batchmode', texfname]
        with open(texpath, 'wt') as ofh:
            ofh.write(contents)
            ofh.flush()
        with open(pdfpath + '.out', 'wb') as logfile:
            p = subprocess.Popen(cmds, cwd=output_dir,
                                 stdout=logfile, stderr=logfile)
            retcode = p.wait()
            p = subprocess.Popen(cmds, cwd=output_dir,
                                 stdout=logfile, stderr=logfile)
            retcode += p.wait()
        if retcode:
            fmtstr = "{}\n returned with exit status {}"
            raise RuntimeError(fmtstr.format(' '.join(cmds), retcode))
        else:
            return pdfpath
    finally:
        if save is True or save == 'True':
            pass
        else:
            if save is False or save == 'False':
                if created_tempdir:
                    shutil.rmtree(output_dir)
            else:
                # interpret path to copy pdf to.
                if not os.path.samefile(pdfpath, save):
                    shutil.copy(pdfpath, save)


def rsys2tablines(rsys, rref0=1, coldelim=' & ',
                  tex=True, ref_fmt=None,
                  unit_registry=None, unit_fmt='{}', k_fmt='%.4g'):
    """
    Generates a table representation of a ReactionSystem.

    Parameters
    ----------
    rsys : ReactionSystem
    rref0 : integer
        default start of index counter (default: 1)
    coldelim : string
        column delimiter (default: ' & ')
    tex : bool
        use latex formated output (default: True)
    ref_fmt : string or callable
        format string of ``ref`` attribute of reactions
    unit_registry : unit registry
        optional (default: None)
    """
    if ref_fmt is None:
        def _doi(s):
            return r'\texttt{\href{http://dx.doi.org/'+s+'}{doi:'+s+'}}'

        def ref_fmt(s):
            if s is None:
                return 'None'
            if tex:
                if isinstance(s, dict):
                    return _doi(s['doi'])
                if s.startswith('doi:'):
                    return _doi(s[4:])
            return s

    def _wrap(s):
        if tex:
            return '\\ensuremath{' + s + '}'
        else:
            return s

    lines = []
    for ri, rxn in enumerate(rsys.rxns):
        rxn_ref = rxn.ref
        if isinstance(rxn.param, RadiolyticBase):
            if unit_registry is not None:
                kunit = get_derived_unit(unit_registry, 'radiolytic_yield')
                k = k_fmt % to_unitless(rxn.param.args[0], kunit)
                k_unit_str = (kunit.dimensionality.latex.strip('$') if tex
                              else kunit.dimensionality)
        else:
            if unit_registry is not None:
                kunit = (get_derived_unit(unit_registry,
                                          'concentration')**(1-rxn.order()) /
                         get_derived_unit(unit_registry, 'time'))
                try:
                    k = k_fmt % to_unitless(rxn.param, kunit)
                    k_unit_str = (kunit.dimensionality.latex.strip('$') if tex
                                  else kunit.dimensionality)
                except Exception:
                    k, k_unit_str = rxn.param.equation_as_string(k_fmt, tex)
            else:
                k_unit_str = '-'
                if isinstance(k_fmt, str):
                    k = k_fmt % rxn.param
                else:
                    k = k_fmt(rxn.param)
        latex_kw = dict(with_param=False)
        if tex:
            latex_kw['substances'] = rsys.substances
            latex_kw['Reaction_around_arrow'] = ('}}' + coldelim + '\\ensuremath{{',
                                                 '}}' + coldelim + '\\ensuremath{{')
        else:
            latex_kw['Reaction_around_arrow'] = (coldelim,)*2
            latex_kw['Reaction_arrow'] = '->'
        lines.append(coldelim.join([
            str(rref0+ri),
            ('\\ensuremath{%s}' if tex else '%s') % latex(rxn, **latex_kw),
            _wrap(k),
            unit_fmt.format(_wrap(k_unit_str)),
            ref_fmt(rxn_ref) if callable(ref_fmt) else ref_fmt.format(rxn_ref)
        ]))

    return lines


def rsys2table(rsys, table_template=None, table_template_dict=None,
               param_name='Rate constant', **kwargs):
    r"""
    Renders user provided table_template with table_template_dict which
    also has 'body' entry generated from `rsys2tablines`.

    Defaults is LaTeX table requiring booktabs package to be used
    (add \usepackage{booktabs} to preamble).

    Parameters
    ----------
    rsys : ReactionSystem
    table_template : string
    table_tempalte_dict : dict used to render table_template (excl. "body")
    param_name : str
        Column header for parameter column
    longtable : bool
        use longtable in defaults. (default: False)
    **kwargs :
        passed onto rsys2tablines

    """
    siunitx = kwargs.pop('siunitx', False)
    line_term = r' \\'
    defaults = {
        'table_env': 'longtable' if kwargs.pop(
            'longtable', False) else 'table',
        'alignment': 'llllSll' if siunitx else 'lllllll',
        'header': kwargs.get('coldelim', ' & ').join([
            'Id.', 'Reactants', '', 'Products', '{%s}' % param_name,
            'Unit', 'Ref'
        ]) + line_term,
        'short_cap': rsys.name,
        'long_cap': rsys.name,
        'label': (rsys.name or 'None').lower()
    }

    if table_template_dict is None:
        table_template_dict = defaults
    else:
        for k, v in defaults:
            if k not in table_template_dict:
                table_template_dict[k] = v

    if 'body' in table_template_dict:
        raise KeyError("There is already a 'body' key in table_template_dict")
    if 'k_fmt' not in kwargs:
        kwargs['k_fmt'] = r'\num{%.4g}' if siunitx else '%.4g'
    table_template_dict['body'] = (line_term + '\n').join(rsys2tablines(
        rsys, **kwargs)
    ) + line_term

    if table_template is None:
        if table_template_dict['table_env'] == 'longtable':
            table_template = tex_templates['table']['longtable']
        else:
            table_template = tex_templates['table']['default']

    return table_template % table_template_dict


def rsys2pdf_table(rsys, output_dir=None, doc_template=None,
                   doc_template_dict=None, save=True, landscape=False,
                   **kwargs):
    """
    Convenience function to render a ReactionSystem as
    e.g. a pdf using e.g. pdflatex.

    Parameters
    ----------
    rsys : ReactionSystem
    output_dir : path to output directory
        (default: system's temporary folder)
    doc_template : string
        LaTeX boiler plate temlpate including preamble,
        document environment etc.
    doc_template_dict : dict (string -> string)
        dict used to render temlpate (excl. 'table')
    longtable : bool
        use longtable in defaults. (default: False)
    **kwargs :
        passed on to `rsys2table`
    """
    if doc_template is None:
        doc_template = tex_templates['document']['default']
    lscape = ['pdflscape' if landscape == 'pdf' else 'lscape'] if landscape else []
    _pkgs = [
        'booktabs', 'amsmath', ('pdftex,colorlinks,unicode=True', 'hyperref')
    ] + lscape
    if kwargs.get('longtable', False):
        _pkgs += ['longtable']
    if kwargs.get('siunitx', False):
        _pkgs += ['siunitx']
    _envs = ['tiny'] + (['landscape'] if landscape else [])
    defaults = {
        'usepkg': '\n'.join([(r'\usepackage' + (
            '[%s]' if isinstance(pkg, tuple) else '') + '{%s}') % pkg for pkg in _pkgs]),
        'begins': '\n'.join([r'\begin{%s}' % env for env in _envs]),
        'ends': '\n'.join([r'\end{%s}' % env for env in _envs[::-1]])
    }

    if doc_template_dict is None:
        doc_template_dict = defaults
    else:
        for k, v in defaults:
            if k not in doc_template_dict:
                doc_template_dict[k] = v

    if 'table' in doc_template_dict:
        raise KeyError("There is already a 'table' key in doc_template_dict")
    doc_template_dict['table'] = rsys2table(rsys, **kwargs)

    contents = doc_template % doc_template_dict

    if isinstance(save, str) and save.endswith('.pdf'):
        texfname = save.rstrip('.pdf') + '.tex'
        pdffname = save
    else:
        texfname = 'output.tex'
        pdffname = 'output.pdf'
    return render_tex_to_pdf(contents, texfname, pdffname, output_dir, save)



# -*- coding: utf-8 -*-
from __future__ import (absolute_import, division, print_function)

from functools import reduce
import math
from operator import add

from ._expr import Expr


def _eval_poly(x, offset, coeffs, reciprocal=False):
    _x0 = x - offset
    _x = _x0/_x0
    res = None
    for coeff in coeffs:
        if res is None:
            res = coeff*_x
        else:
            res += coeff*_x

        if reciprocal:
            _x /= _x0
        else:
            _x *= _x0
    return res


def _mk_Poly(parameter_name, reciprocal=False, shift_name='shift'):
    """ Class factory of Expr subclass for (shifted) polynomial

    Parameters
    ----------
    parameter: str
        name of paramter
    reciprocal: bool
        whether the polynomial is in the reciprocal of the parameter

    Returns
    -------
    Expr subclass for a shifted polynomial with the args: offset, p0, p1, ...
    the class has the method "eval_poly" with same signature as __call__


    Examples
    --------
    >>> P = _mk_Poly('x')
    >>> p = P([3, 5, 7, 2])
    >>> p.eval_poly({'x': 13}) == 5 + 7*(13-3) + 2*(13-3)**2
    True

    """
    class Poly(Expr):
        """ Args: shift, p0, p1, ... """
        argument_names = (shift_name, Ellipsis)
        parameter_keys = (parameter_name,)
        skip_poly = 0

        def eval_poly(self, variables, backend=math):
            all_args = self.all_args(variables, backend=backend)
            x = variables[parameter_name]
            offset, coeffs = all_args[self.skip_poly], all_args[self.skip_poly+1:]
            return _eval_poly(x, offset, coeffs, reciprocal)
    return Poly


def _mk_PiecewisePoly(parameter, reciprocal=False):
    """ Class factory of Expr subclass for piecewise (shifted) polynomial """
    class PiecewisePoly(Expr):
        """ Args: npolys, ncoeff0, lower0, upper0, ncoeff1, ..., shift0, p0_0, p0_1, ... shiftn, p0_n, p1_n, ... """
        argument_names = ('npolys', Ellipsis)
        parameter_keys = (parameter,)
        skip_poly = 0

        def eval_poly(self, variables, backend=math):
            all_args = self.all_args(variables, backend=backend)[self.skip_poly:]
            npoly = all_args[0]
            arg_idx = 1
            poly_args = []
            meta = []
            for poly_idx in range(npoly):
                meta.append(all_args[arg_idx:arg_idx+3])  # nargs, lower, upper
                arg_idx += 3
            for poly_idx in range(npoly):
                narg = 1+meta[poly_idx][0]
                poly_args.append(all_args[arg_idx:arg_idx+narg])
                arg_idx += narg
            if arg_idx != len(all_args):
                raise Exception("Bug in PiecewisePoly.eval_poly")

            x = variables[parameter]
            try:
                pw = backend.Piecewise
            except AttributeError:
                for (ncoeff, lower, upper), args in zip(meta, poly_args):
                    if lower <= x <= upper:
                        return _eval_poly(x, args[0], args[1:], reciprocal)
                else:
                    raise ValueError("not within any bounds: %s" % str(x))
            else:
                return pw(*[(_eval_poly(x, a[0], a[1:], reciprocal),
                             backend.And(l <= x, x <= u)) for (n, l, u), a in zip(meta, poly_args)])

        @classmethod
        def from_polynomials(cls, bounds, polys, inject=[], **kwargs):
            if any(p.parameter_keys != (parameter,) for p in polys):
                raise ValueError("Mixed parameter_keys")
            npolys = len(polys)
            if len(bounds) != npolys:
                raise ValueError("Length mismatch")

            meta = reduce(add, [[len(p.args[p.skip_poly:]) - 1, l, u] for (l, u), p in zip(bounds, polys)])
            p_args = reduce(add, [p.args[p.skip_poly:] for p in polys])
            return cls(inject + [npolys] + meta + p_args, **kwargs)
    return PiecewisePoly



# -*- coding: utf-8 -*-
from __future__ import (absolute_import, division, print_function)

import pytest

from .._solution import Solution, QuantityDict
from ..util.testing import requires
from ..units import magnitude, units_library, to_unitless, default_units as u


@requires(units_library)
def test_QuantityDict():
    # QuantityDict * scalar_quantity
    c = QuantityDict(u.molar, {})
    c['H2O'] = 55.4 * u.molar
    with pytest.raises(ValueError):
        c['HCl'] = 3 * u.kg

    with pytest.raises(ValueError):
        QuantityDict(u.molar, {'a': u.mole})

    V = .4*u.dm3
    n = c*V
    assert isinstance(n, QuantityDict)

    # For the following to work major changes to quantities needs to be made:
    # n = V*c
    # assert isinstance(n, QuantityDict)

    assert n.isclose({'H2O': 55.4*.4*u.mol})
    assert abs(to_unitless(n['H2O'], u.mol) - 55.4*.4) < 1e-14

    c2 = c.rescale(u.mol/u.cm3)
    assert abs(magnitude(c2['H2O']) - 55.4e-3) < 1e-6


def _get_s1_s2():
    s1 = Solution(0.1*u.dm3, {'CH3OH': 0.1 * u.molar})
    s2 = Solution(0.3*u.dm3, {'CH3OH': 0.4 * u.molar, 'Na+': 2e-3*u.molar,
                              'Cl-': 2e-3*u.molar})
    return s1, s2


@requires(units_library)
def test_Solution__add():
    s1, s2 = _get_s1_s2()
    s3 = s1 + s2
    assert abs(to_unitless(s3.volume - 4e-4 * u.m**3, u.dm3)) < 1e-15


@requires(units_library)
def test_Solution__isclose():
    s1, s2 = _get_s1_s2()
    s3 = s1 + s2
    assert s3.concentrations.isclose({'CH3OH': 0.325*u.molar, 'Na+': 1.5e-3*u.molar, 'Cl-': 1.5e-3*u.molar})


@requires(units_library)
def test_Solution__dissolve():
    s1, s2 = _get_s1_s2()
    s4 = (s1 + s2).dissolve({'CH3OH': 1*u.gram})
    mw = (12.011 + 4*1.008 + 15.999)
    assert abs(s4.concentrations['CH3OH'] - (0.325 + 1/mw/.4)*u.molar) < 1e-7


@requires(units_library)
def test_Solution__withdraw():
    s1, s2 = _get_s1_s2()
    s3 = (s1 + s2)
    s4 = s3.withdraw(.2 * u.dm3)
    assert s4 == s3
    assert s4 == (s1 + s2).withdraw(.2 * u.dm3)
    assert s4 != s1 + s2
